import re
from flask import request, jsonify
from .models import Base
from .controller import Controller
from .exceptions import UnauthorizedError, NoTargetError, InvalidTargetError, InvalidTypeError
import jwt
import os
from .models import User


class Router:
    def __init__(self, app, db):
        # Register models
        self.app = app
        self.db = db
        self.tree = {}
        self.register_base_routes(Controller)
        self.register_model(Base)

    def register_model(self, superclass):
        for M in superclass.__subclasses__():
            self.register_routes_for_model(self.to_snake(M.__name__), M)
            self.register_model(M)

    def register_routes_for_model(self, target, M):
        self.tree[target] = {}
        c = getattr(M, 'Controller', Controller)
        routes = []
        for m_name in dir(c):
            method = getattr(c, m_name)

            if getattr(method, 'is_route', False):
                routes.append(m_name)

            if getattr(method, 'is_http_route', False):
                url = '/api/{}/{}/'.format(target, m_name)
                self.add_http_route(url, c, method, target)

        self.tree[target]['routes'] = routes
        self.tree[target]['Controller'] = c
        self.tree[target]['Model'] = M

    def add_http_route(self, url, C, route, target):
        def wrapped_route():
            return self.route_http(request, C, route, target)

        options = getattr(route, 'options')
        endpoint = options.pop('endpoint', route.__name__)
        self.app.add_url_rule(url, endpoint, wrapped_route, **options)

    def to_snake(self, name):
        s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
        return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()

    def register_base_routes(self, c):
        self.base_routes = []

        for m_name in dir(c):
            method = getattr(c, m_name)

            if getattr(method, 'is_targetless', False):
                self.base_routes.append(m_name)

    def route_http(self, req, C, route, target):
        body = {}
        body['authorization'] = req.args.get('authorization')
        auth = self.check_auth(body)
        body['target'] = target
        M = self.tree[target]['Model']
        body['type'] = route.__name__

        try:
            c = C(self.db, body, auth, request=request)
            method = getattr(c, route.__name__)

            self.require_auth(c, method, auth)
            res = method(M)
        except Exception as e:
            if isinstance(e, UnauthorizedError) or not auth:
                response = jsonify({'message': UnauthorizedError.message})
                response.status_code = 403
            # elif isinstance(e, ArchonError):
            #     response = jsonify({'message', e.message})
            #     response.status_code = 412
            else:
                response = jsonify({'message': str(e)})
                response.status_code = 500
            return response

        return jsonify(res)

    def check_auth(self, body):
        currentUser = None
        if 'authorization' not in body:
            return False

        try:
            userData = jwt.decode(body['authorization'], os.environ.get('CY_SECRET_KEY', ''), algorithms=['HS256'])
            currentUser = User.query.get(userData['id'])
        except jwt.InvalidTokenError:
            return False

        return currentUser

    def route(self, connection, body, auth):
        # Handle base route (no target needed)
        if body['type'] in self.base_routes:
            c = Controller(self.db, body, auth, connection=connection)
            method = getattr(c, body['type'])
            self.require_auth(c, method, auth)
            return method()

        if 'target' not in body:
            raise NoTargetError()

        t_name = body['target']
        if t_name not in self.tree.keys():
            raise InvalidTargetError()

        target = self.tree[t_name]['Model']
        C = self.tree[t_name]['Controller']
        c = C(self.db, body, auth, connection=connection)
        method = getattr(c, body['type'], None)

        if not method or not getattr(method, 'is_route', False):
            raise InvalidTypeError()

        self.require_auth(c, method, auth)
        # Call the method with the class as param
        return method(target)

    def require_auth(self, controller, method, auth):
        public = getattr(method, 'is_public', False)
        if not public and not auth:
            raise UnauthorizedError()



import json
from .testapp.app import app, db
from archon.test import LoggedInTestCase, Client, MockWebSocket


save_company = {
    'target': 'company',
    'type': 'save',
    'data': {
        'name': 'Snackbar',
    },
}

subscribe_company = {
    'requestId': '1234',
    'target': 'company',
    'type': 'subscribe',
}


class TestSave(LoggedInTestCase):
    def setUp(self):
        self.client = Client(app, db)
        super().setUp()

    def test_save_model(self):
        ws = MockWebSocket()
        ws.mock_incoming_message(save_company)

        self.client.open_connection(ws)

        self.assertEqual(1, len(ws.outgoing_messages))
        m = ws.outgoing_messages[0]

        self.assertDictEqual({
            'type': 'save',
            'target': 'company',
            'code': 'success',
            'data': {
                'id': 1,
                'name': 'Snackbar'
            }
        }, json.loads(m))


class TestSub(LoggedInTestCase):
    def setUp(self):
        self.client = Client(app, db)
        super().setUp()

    def test_save(self):
        ws = MockWebSocket()

        ws.mock_incoming_message(subscribe_company)
        ws.mock_incoming_message(save_company)

        self.client.open_connection(ws)

        self.assertEqual(3, len(ws.outgoing_messages))
        r1 = json.loads(ws.outgoing_messages[0])
        r2 = json.loads(ws.outgoing_messages[1])
        r3 = json.loads(ws.outgoing_messages[2])

        self.assertEqual('publish', r1['type'])
        self.assertEqual('success', r1['code'])

        self.assertEqual('save', r3['type'])
        self.assertEqual('success', r3['code'])

        self.assertDictEqual({
            'type': 'publish',
            'target': 'company',
            'requestId': '1234',
            'data': {
                'add': [{
                    'id': 1,
                    'name': 'Snackbar',
                }]
            },
        }, r2)



import requests
import json
import base64
from .testapp.app import app, db
from archon.test import TestCase, Client, MockResponse, MockWebSocket
from archon.models import User
from greenlet import greenlet


subscribe_company = {
    'target': 'company',
    'type': 'subscribe',
}

authenticate = {
    'type': 'authenticate',
    'data': {
        'code': 'foo',
    },
}


def create_fake_google_id_token():
    u_data = {
        'name': 'Henk de Vries',
        'email': 'henk@devries.nl',
        'picture': 'bla',
    }

    body_decoded = json.dumps(u_data)
    body = base64.urlsafe_b64encode(body_decoded.encode())
    # Strip the = from the body to reproduce google's twisted response
    body_malformed = body.decode().split('=')[0]
    return 'foo.{}.bar'.format(body_malformed)


class TestAuth(TestCase):
    def setUp(self):
        self.client = Client(app, db)
        super().setUp()

    def test_unauth_request(self):
        ws = MockWebSocket()
        ws.mock_incoming_message(subscribe_company)
        self.client.open_connection(ws)

        self.assertEqual(1, len(ws.outgoing_messages))
        m = ws.outgoing_messages[0]

        self.assertDictEqual({
            'type': 'subscribe',
            'message': 'Unauthorized',
            'code': 'error',
        }, json.loads(m))

    def test_google_oauth(self):
        def mock_oauth(request, **kwargs):
            return MockResponse({
                'id_token': create_fake_google_id_token()
                }, 200)

        self.client.set_mock_api(mock_oauth)
        ws = MockWebSocket()
        ws.mock_incoming_message(authenticate)

        g = greenlet(self.client.open_connection)
        g.switch(ws, app=self.client.app)

        self.assertEqual(1, requests.post.call_count)

        call_params = requests.post.call_args_list[0][1]['params']
        self.assertTrue('client_id' in call_params)
        self.assertTrue('client_secret' in call_params)
        self.assertTrue('redirect_uri' in call_params)
        self.assertTrue(call_params['code'], authenticate['data']['code'])
        self.assertTrue(call_params['grant_type'], 'authorization_code')

        users = self.client.db.session.query(User).all()
        self.assertEqual(1, len(users))
        self.assertEqual('henk@devries.nl', users[0].email)

        self.assertEqual(1, len(ws.outgoing_messages))
        m = json.loads(ws.outgoing_messages[0])
        self.assertEqual('authenticate', m['type'])
        self.assertEqual('success', m['code'])
        self.assertTrue('authorization' in m)

        # Now check that the response auth code actually works
        req_bootstrap = {
            'type': 'bootstrap',
            'authorization': m['authorization'],
        }

        ws.mock_incoming_message(req_bootstrap)
        g.switch()

        self.assertEqual(2, len(ws.outgoing_messages))
        m = json.loads(ws.outgoing_messages[1])
        self.assertEqual('bootstrap', m['type'])
        self.assertEqual('success', m['code'])
        self.assertEqual('Henk de Vries', m['data']['username'])


class TestRecon(TestCase):
    '''
    In the Router we have some sanity checks
    These sanitychecks may only return useful information
    if the user is authenticated
    '''
    def setUp(self):
        self.client = Client(app, db)
        super().setUp()

    def test_no_target(self):
        ws = MockWebSocket()
        ws.mock_incoming_message({
            'type': 'subscribe',
        })
        self.client.open_connection(ws)

        self.assertEqual(1, len(ws.outgoing_messages))
        m = ws.outgoing_messages[0]

        self.assertDictEqual({
            'type': 'subscribe',
            'message': 'Unauthorized',
            'code': 'error',
        }, json.loads(m))

    def test_invalid_target(self):
        ws = MockWebSocket()
        ws.mock_incoming_message({
            'type': 'subscribe',
            'target': "Robert';) DROP TABLE Students; --"
        })
        self.client.open_connection(ws)

        self.assertEqual(1, len(ws.outgoing_messages))
        m = ws.outgoing_messages[0]

        self.assertDictEqual({
            'type': 'subscribe',
            'message': 'Unauthorized',
            'code': 'error',
        }, json.loads(m))

    def test_invalid_type(self):
        ws = MockWebSocket()
        ws.mock_incoming_message({
            'type': 'PUT',
            'target': 'company',
        })
        self.client.open_connection(ws)

        self.assertEqual(1, len(ws.outgoing_messages))
        m = ws.outgoing_messages[0]

        self.assertDictEqual({
            'type': 'PUT',
            'message': 'Unauthorized',
            'code': 'error',
        }, json.loads(m))



from archon.models import db, Base


class Employee(Base, db.Model):
    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String(200))

    company_id = db.Column(db.Integer, db.ForeignKey('company.id'))
    company = db.relationship('Company', backref=db.backref('employees', lazy='dynamic'))

    started_at = db.Column(db.DateTime)
    birthday = db.Column(db.Date)



import json
from .testapp.app import app, db
from .testapp.models import Company
from archon.test import LoggedInTestCase, Client, MockWebSocket

save_company = {
    'target': 'company',
    'type': 'save',
    'data': {
        'name': 'Snackbar',
    },
}

subscribe_company = {
    'requestId': '1234',
    'target': 'company',
    'type': 'subscribe',
}


class TestSubscribe(LoggedInTestCase):
    def setUp(self):
        self.client = Client(app, db)
        super().setUp()

    def test_publishes_existing_collection_upon_subscription(self):
        ws = MockWebSocket()
        ws.mock_incoming_message(subscribe_company)

        c = Company({
                'name': 'Slagerij Henk',
            })
        db.session.add(c)
        db.session.commit()

        self.client.open_connection(ws)

        self.assertEqual(1, len(ws.outgoing_messages))
        m = ws.outgoing_messages[0]

        self.assertDictEqual({
            'type': 'publish',
            'target': 'company',
            'requestId': '1234',
            'data': {
                'add': [{
                    'id': c.id,
                    'name': c.name,
                }],
                'update': [],
                'delete': [],
            },
            'code': 'success',
        }, json.loads(m))



class ArchonError(Exception):
    message = 'Something went wrong'


class InvalidTargetError(ArchonError):
    message = 'Invalid target given'


class NoTargetError(ArchonError):
    message = 'No target given'


class UnauthorizedError(ArchonError):
    message = 'Unauthorized'


class InvalidTypeError(ArchonError):
    message = 'Invalid type given'



import json
import os
import requests
import copy
import base64
from .models import User


def targetless_route(func):
    func.is_targetless = True
    func.is_route = True
    return func


def public_route(func):
    func.is_public = True
    return func


def model_route(func):
    func.is_route = True
    return func


def http_route(**options):
    def decorator(func):
        func.is_http_route = True
        func.options = options
        return func
    return decorator


class Controller:
    db = None
    connection = None
    body = None
    currentUser = None

    def __init__(self, db, body, currentUser, connection=None, request=None):
        self.db = db
        self.connection = connection
        self.request = request
        self.body = body
        self.currentUser = currentUser

    def error(self, msg):
        return {
            'type': self.body['type'],
            'code': 'error',
            'message': msg if msg else '',
        }

    @targetless_route
    def bootstrap(self):
        output = copy.copy(self.currentUser.dump())
        return {
            'type': self.body['type'],
            'code': 'success',
            'data': output,
        }

    @model_route
    def save(self, cls):
        data = self.body['data']

        # Create instance if id is not given
        # Ignore binary post data (file upload)
        if not type(data) == bytes and 'id' in data and data['id'] is not None:
            return self.error('ID given when saving, try using type=update')

        try:
            m = cls(data, self.currentUser)
        except Exception as e:
            return self.error(str(e))

        self.db.session.add(m)
        self.db.session.commit()

        result = m.dump()
        return {
            'type': self.body['type'],
            'target': self.body['target'],
            'code': 'success',
            'data': result,
        }

    @model_route
    def update(self, cls):
        data = self.body['data']

        # Create instance if id is not given
        if 'id' not in data or data['id'] is None:
            return self.error('No id given')

        m = self.db.session.query(cls).get(data['id'])
        mSnapshot = m.dump()
        m.parse(data, self.currentUser, 'update')

        self.db.session.add(m)
        self.db.session.commit()

        result = m.dump()
        return {
            'type': self.body['type'],
            'target': self.body['target'],
            'snapshot': mSnapshot,
            'code': 'success',
            'data': result,
        }

    @model_route
    def delete(self, cls):
        data = self.body.get('data')
        if not data:
            return self.error('No data given')

        # Create instance if id is not given
        if 'id' not in data or data['id'] is None:
            return self.error('No id given')

        m = self.db.session.query(cls).get(data['id'])
        self.db.session.delete(m)
        self.db.session.commit()

        result = m.dump()
        return {
            'type': self.body['type'],
            'target': self.body['target'],
            'code': 'success',
            'requestId': self.body.get('requestId', None),
            'data': result,
        }

    @model_route
    def subscribe(self, cls):
        scope = self.body['data'] if 'data' in self.body else {}

        result = cls.find(self.db.session, scope)

        # Mark the socket as subscribing so we know what is listening to what
        self.connection.subscribe(self.body['requestId'], self.body['target'], scope)

        if 'requestId' not in self.body:
            return self.error('No requestId given')

        return {
            'type': 'publish',
            'target': self.body['target'],
            'code': 'success',
            'requestId': self.body['requestId'],
            'data': {
                'add': result.dump(),
                'update': [],
                'delete': [],
            }
        }

    @targetless_route
    def unsubscribe(self):
        reqId = self.body.get('requestId', None)

        if not reqId:
            return self.error('No requestId given')

        success = self.connection.unsubscribe(reqId)
        if not success:
            return self.error('Invavlid requestId given')

        return {
            'type': 'unsubscribe',
            'code': 'success',
            'requestId': self.body['requestId'],
        }

    @public_route
    @targetless_route
    def authenticate(self):
        data = {
            'client_id': os.environ.get('CY_OAUTH_CLIENT_ID'),
            'client_secret': os.environ.get('CY_OAUTH_CLIENT_SECRET'),
            'redirect_uri': os.environ.get('CY_REDIRECT_URI'),
            'code': self.body['data']['code'],
            'grant_type': 'authorization_code',
        }

        r1 = requests.post(os.environ.get('CY_OAUTH_URL'), params=data)

        if r1.status_code != 200:
            return self.error(r1.json()['error_description'])

        # Alright so for some reason, google returns the client info encapsulated in a JWT token.
        # This JWT token has an invalid signature, so we just b64 decode the body.
        #
        # https://stackoverflow.com/questions/16923931/python-google-ouath-authentication-decode-and-verify-id-token
        token = r1.json()['id_token'].split('.')

        # Repad the b64 body
        body_encoded = token[1]
        body_repadded = body_encoded + '=' * (4 - len(body_encoded) % 4)
        body_decoded = base64.urlsafe_b64decode(body_repadded).decode()
        u_data = json.loads(body_decoded)

        user = self.db.session.query(User).filter(User.email == u_data['email']).first()

        if not user:
            u = {
                'username': u_data['name'],
                'email': u_data['email'],
                'avatar_url': u_data['picture'],
                'display_name': u_data['name'],
            }
            user = User(u)
            self.db.session.add(user)
            self.db.session.commit()

        token = user.create_session()

        return {
            'type': 'authenticate',
            'code': 'success',
            'authorization': token,
        }


# class SocketController(BaseController):
#     db = None
#     requestContainer = None
#     body = None
#     currentUser = None

#     def __init__(self, db, socketContainer, message):
#         super().__init__()
#         self.db = db
#         self.socketContainer = socketContainer
#         self.message = message

#     def handle(self):
#         if self.message == 'ping':
#             # todo keepalive logic
#             return 'pong'

#         self.body = self._parse_body()
#         return self.follow()

#     def _parse_body(self):
#         return json.loads(self.message)


# class RequestController(BaseController):
#     db = None
#     request = None
#     body = None
#     currentUser = None

#     def __init__(self, db, request):
#         super().__init__()
#         self.db = db
#         self.request = request

#     def handle(self):
#         self.body = self._parse_body()
#         return self.follow()

#     def _get_type(self, request):
#         if request.method == 'POST':
#             return 'save'
#         if request.method == 'PUT' or request.method == 'PATCH':
#             return 'update'
#         if request.method == 'DELETE':
#             return 'delete'
#         if request.method == 'GET':
#             return 'get'
#         return False

#     def _parse_body(self):
#         body = {}
#         req = self.request
#         body['authorization'] = req.args.get('authorization')
#         body['target'] = req.endpoint
#         body['type'] = self._get_type(req)
#         if body['type'] == 'save' and not req.data:
#             if not req.files:
#                 self.error('save without data or file detected')
#             body['data'] = {'file': req.files['file']}

#         return body



from archon.models import db, Base
from archon.controller import Controller, model_route, http_route


class CompanyController(Controller):
    @model_route
    def ask_for_raise(self, cls):
        return {
            'answer': 'no',
            'code': 'error',
        }

    @model_route
    def list_employees(self, cls):
        data = self.body['data']
        company = cls.query.get(data['id'])
        return {
            'code': 'success',
            'type': self.body['type'],
            'data': [e.name for e in company.employees],
        }

    @model_route
    def list_employees_error(self, cls):
        # A controller has no type, so this should error
        return {'type': self.type}

    @http_route(methods=['GET'])
    def foo(self, cls):
        return {'foo': 'bar'}

    @http_route(methods=['GET'])
    def foo_error(self, cls):
        # A controller has no type, so this should error
        return {'type': self.type}


class Company(Base, db.Model):
    Controller = CompanyController

    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String(200))



from .testapp.app import app, db
from archon.test import TestCase, Client, MockWebSocket


class TestPing(TestCase):
    def setUp(self):
        self.client = Client(app, db)
        super().setUp()

    def test_pong(self):
        ws = MockWebSocket()

        ws.mock_incoming_message('ping')

        self.client.open_connection(ws)

        self.assertEqual(1, len(ws.outgoing_messages))
        # Whoops our response is a json encoded string
        self.assertEqual('pong', ws.outgoing_messages[0])



import json
from .testapp.app import app, db
from .testapp.models import Employee, Company
from archon.test import LoggedInTestCase, Client, MockWebSocket, greenlet


save_company = {
    'target': 'company',
    'type': 'save',
    'data': {
        'name': 'Snackbar',
    },
}


class TestParse(LoggedInTestCase):
    def setUp(self):
        self.client = Client(app, db)
        super().setUp()

    def test_save_model(self):
        ws = MockWebSocket()
        ws.mock_incoming_message(save_company)
        g = greenlet(self.client.open_connection)
        g.switch(ws, app=self.client.app)

        self.assertEqual(1, len(ws.outgoing_messages))
        c = json.loads(ws.outgoing_messages[0])['data']

        ws.mock_incoming_message({
            'target': 'employee',
            'type': 'save',
            'data': {
                'name': 'Henk',
                'company': c['id'],
                'started_at': '2017-09-16T15:59:36+02:00',
                'birthday': '1993-03-21',
            },
        })
        g.switch()

        self.assertEqual(2, len(ws.outgoing_messages))
        r2 = json.loads(ws.outgoing_messages[1])
        self.assertDictEqual({
            'target': 'employee',
            'type': 'save',
            'code': 'success',
            'data': {
                'id': 1,
                'name': 'Henk',
                'company': c['id'],
                'started_at': '2017-09-16T13:59:36+00:00',
                'birthday': '1993-03-21',
            },
        }, r2)

        c = Company.query.get(1)
        e = Employee.query.get(1)

        self.assertEqual(e.company, c)
        self.assertEqual(c.employees.first(), e)



# from archon.controller import SocketController, RequestController
# from flask import jsonify
import json
import uuid
from .exceptions import ArchonError, UnauthorizedError


class Subscription():
    socket = None
    reqId = None
    target = None
    scope = None

    def __init__(self, socket, reqId, target, scope=None):
        self.socket = socket
        self.reqId = reqId
        self.target = target
        self.scope = scope

    # Check if this subscription
    # matches the scope & target
    def is_in_scope(self, target, item):
        if self.target != target:
            return False
        if self.scope:
            for skey, sval in self.scope.items():
                if skey not in item.keys() or sval != item[skey]:
                    return False

        return True

    def handle_event(self, target, _type, item, snapshot):
        is_subscribed = self.is_in_scope(target, item)

        if _type == 'save':
            if is_subscribed:
                self.publish(item, 'add')
            return
        # is_subscribed (instead of was_subscribed) may be a mindfuck as the model is now deleted,
        # but we dont send a snapshot when deleting, the res.data just is the original model
        if _type == 'delete':
            if is_subscribed:
                self.publish(item, 'remove')
            return

        # _type == 'update'
        was_subscribed = self.is_in_scope(target, snapshot)

        if is_subscribed and not was_subscribed:
            return self.publish(item, 'add')
        if was_subscribed and not is_subscribed:
            return self.publish(item, 'remove')
        if is_subscribed and was_subscribed:
            return self.publish(item, 'update')

        return

    def publish(self, item, publish_type):
        res = json.dumps({
            'type': 'publish',
            'target': self.target,
            'requestId': self.reqId,
            'data': {
                publish_type: [item],
            }
        })

        if self.socket.ws.closed:
            return

        self.socket.ws.send(res)


# class RequestContainer():
#     request = None

#     def __init__(self, hub, request):
#         self.uuid = uuid.uuid4()
#         self.hub = hub
#         self.request = request

#     def handle(self, db):
#         controller = RequestController(db, self)
#         res = controller.handle()

#         if type(res) is dict and res['code'] == 'success':
#             # Handle publish for successful saves, deletes and updates
#             if res['type'] in ['save', 'update', 'delete']:
#                 self.hub.handle_event(res['target'], res['type'], res['data'], res.get('snapshot', None))

#         response = jsonify(res)
#         # TODO {error-handling}
#         if res['code'] == 'error':
#             response.status_code = 400

#         return response


class Connection():
    # This only exists because
    # I want to do some pubsub scoping logic
    # And it doesnt belong in the controller
    hub = None
    ws = None

    def __init__(self, hub, ws):
        self.uuid = uuid.uuid4()
        self.hub = hub
        self.subs = []
        self.ws = ws

    def subscribe(self, requestId, target, scope=None):
        s = Subscription(self, requestId, target, scope)
        self.subs.append(s)

    def unsubscribe(self, requestId):
        s = None
        for sub in self.subs:
            if sub.reqId == requestId:
                s = sub

        if not s:
            return False

        self.subs.remove(s)
        return True

    def handle_event(self, target, _type, item, snapshot):
        if self.ws.closed:
            self.hub.remove(self)
            return

        for sub in self.subs:
            sub.handle_event(target, _type, item, snapshot)

    def handle(self, router, message):
        if message == 'ping':
            self.ws.send('pong')
            return

        body = json.loads(message)
        auth = router.check_auth(body)

        # Let the router route the request
        # to the correct type/target
        try:
            res = router.route(self, body, auth)
        except Exception as e:
            msg = UnauthorizedError.message

            if auth:
                if isinstance(e, ArchonError):
                    msg = e.message
                else:
                    msg = str(e)

            res = {
                'type': body['type'],
                'code': 'error',
                # Make sure we don't leak info to unauthorized users
                'message': msg
            }

        if type(res) is dict and res['code'] == 'success':
            # Handle publish for successful saves, deletes and updates
            if res['type'] in ['save', 'update', 'delete']:
                self.hub.handle_event(res['target'], res['type'], res['data'], res.get('snapshot', None))

        self.ws.send(json.dumps(res))


class Hub():

    def __init__(self):
        self.sockets = []

    def handle_event(self, target, _type, item, snapshot):
        # Clone self.sockets because they may be removed when closed
        for socket in list(self.sockets):
            socket.handle_event(target, _type, item, snapshot)

    def add(self, ws):
        socket = Connection(self, ws)
        self.sockets.append(socket)
        return socket

    def remove(self, socket):
        self.sockets.remove(socket)



from .testapp.app import app, db
from archon.test import LoggedInTestCase, TestCase, Client
import json


class TestUnAuth(TestCase):
    def setUp(self):
        self.client = Client(app, db)
        super().setUp()

    def test_unauthorized(self):
        res = self.client.http_client.get('/api/company/foo/')
        self.assertEqual(403, res.status_code)
        self.assertEqual(json.loads(res.data.decode()), {
            'message': 'Unauthorized'})


class TestHttp(LoggedInTestCase):
    # TODO: find a way to extend the flask client get...
    def setUp(self):
        self.client = Client(app, db)
        super().setUp()

    def test_custom_http_route(self):
        res = self.client.http_client.get(
            '/api/company/foo/',
            query_string={
                'authorization': self.client.auth_token
                })

        self.assertEqual(json.loads(res.data.decode()), {
            'foo': 'bar'})

    def test_uncaught_exception(self):
        res = self.client.http_client.get(
            '/api/company/foo_error/',
            query_string={
                'authorization': self.client.auth_token
                })

        self.assertEqual(500, res.status_code)
        self.assertEqual(json.loads(res.data.decode()), {
            'message': "'CompanyController' object has no attribute 'type'"})



from dateutil import parser, tz
from sqlalchemy.types import DateTime, Date, Enum
from flask_sqlalchemy import SQLAlchemy
import datetime
import jwt
import os

db = SQLAlchemy()


class Collection():
    def __init__(self, models=[]):
        self.models = models

    def __len__(self):
        return len(self.models)

    def dump(self):
        return [m.dump() for m in self.models]


class Base:
    def __init__(self, data={}, currentUser=None):
        self.parse(data, currentUser, 'save')

    def parse(self, data, currentUser=None, reqType=None):
        for col in self.__table__.columns:
            key = col.name

            # If it is a foreign key, we check the input for the key without `_id`
            # EG if we are at col 'project_id' and it is a fk, we check the input for `project`
            if len(col.foreign_keys):
                assert key.endswith('_id')
                # Split the keyname in _, throw away the last part (_id) and join the rest
                key = '_'.join(key.split('_')[:-1])
                if key in data:
                    setattr(self, key + '_id', data[key])
                    continue

            if key in data:
                if type(col.type) == DateTime and data[key] is not None:
                    d = parser.parse(data[key])                                   # Convert to python datetime
                    data[key] = d.astimezone(tz.gettz('UTC')) if d.tzinfo else d  # Convert to UTC

                if type(col.type) == Date and not isinstance(data[key], datetime.date) and data[key] is not None:
                    data[key] = parser.parse(data[key]).date()

                if type(col.type) == Enum and data[key] is not None:
                    data[key] = data[key].value

                setattr(self, key, data[key])

    def __repr__(self):
        return '<Model %r>' % self.id

    def dump(self):
        data = {}
        for col in self.__table__.columns:
            key = col.name
            val = getattr(self, key)

            # Return {'project': id} instead of {'project_id': id}
            if len(col.foreign_keys):
                assert key.endswith('_id')
                key = '_'.join(key.split('_')[:-1])

            if type(col.type) == DateTime and val is not None:
                # Make aware and format as iso
                val = val.replace(tzinfo=tz.gettz('UTC')).isoformat()

            if type(col.type) == Date and val is not None:
                val = val.isoformat()

            data[key] = val
        return data

    @classmethod
    def find(cls, session, scope):
        query = session.query(cls)

        for col in cls.__table__.columns:
            dbKey = col.name
            scopeKey = dbKey

            # Translate 'project_id' to 'project', a relation key shorthand
            if len(col.foreign_keys):
                assert dbKey.endswith('_id')
                scopeKey = '_'.join(dbKey.split('_')[:-1])

            if dbKey in scope or scopeKey in scope:
                val = scope[scopeKey]
                query = query.filter_by(**{dbKey: val})
                continue

        return Collection(query.all())


class User(Base, db.Model):
    __tablename__ = 'users'

    id = db.Column(db.Integer, primary_key=True)
    email = db.Column(db.String(100))
    username = db.Column(db.String(50))
    display_name = db.Column(db.String(50))
    avatar_url = db.Column(db.String(300))

    def create_session(self):
        secret = os.environ.get('CY_SECRET_KEY', '')
        payload = self.dump()
        payload['exp'] = datetime.datetime.utcnow() + datetime.timedelta(weeks=8)
        return jwt.encode(payload, secret, algorithm='HS256').decode('utf-8')



from flask import Flask
from flask_sockets import Sockets


def create_app(settings=None):
    app = Flask(__name__)

    if settings:
        app.config.from_object(settings)
    else:
        app.config['SQLALCHEMY_DATABASE_URI'] = 'sqlite://'

    app.config['SQLALCHEMY_TRACK_MODIFICATIONS'] = False
    sockets = Sockets(app)

    from archon.models import db
    db.init_app(app)

    from archon.hub import Hub
    from archon.router import Router
    app.hub = Hub()
    app.router = Router(app, db)

    @sockets.route('/api/')
    def open_socket(ws):
        connection = app.hub.add(ws)
        while not connection.ws.closed:
            message = connection.ws.receive()
            if message:
                connection.handle(app.router, message)

    return app, db



import json
from .testapp.app import app, db
from archon.test import LoggedInTestCase, Client, MockWebSocket, greenlet

save_company = {
    'target': 'company',
    'type': 'save',
    'data': {
        'name': 'Garage',
    },
}

update_company = {
    'target': 'company',
    'type': 'update',
    'data': {
        'name': 'Nailsalon',
        'id': 1,
    },
}

subscribe_company = {
    'requestId': '1234',
    'target': 'company',
    'type': 'subscribe',
}

unsubscribe_company = {
    'requestId': '1234',
    'type': 'unsubscribe',
}


class TestUnsubscribe(LoggedInTestCase):
    def setUp(self):
        self.client = Client(app, db)
        super().setUp()

    def test_stop_receiving_publishes(self):
        ws1 = MockWebSocket()
        ws2 = MockWebSocket()
        ws1.mock_incoming_message(subscribe_company)

        g1 = greenlet(self.client.open_connection)
        g1.switch(ws1, app=self.client.app)

        ws2.mock_incoming_message(save_company)
        g2 = greenlet(self.client.open_connection)
        g2.switch(ws2, app=self.client.app)

        # check ws1 receives publish
        self.assertEqual(2, len(ws1.outgoing_messages))
        self.assertDictEqual({
            'type': 'publish',
            'target': 'company',
            'requestId': '1234',
            'data': {'add': [{
                'id': 1,
                'name': 'Garage'}]}
            }, json.loads(ws1.outgoing_messages[1]))

        ws1.mock_incoming_message(unsubscribe_company)
        g1.switch()

        ws2.mock_incoming_message(update_company)
        g2.switch()
        self.assertEqual(2, len(ws2.outgoing_messages))
        self.assertEqual('success', json.loads(ws2.outgoing_messages[1])['code'])

        # check ws1 does no receives publish
        self.assertEqual(3, len(ws1.outgoing_messages))
        self.assertDictEqual({
            'type': 'unsubscribe',
            'code': 'success',
            'requestId': '1234',
            }, json.loads(ws1.outgoing_messages[2]))



import json
from .testapp.app import app, db
from .testapp.models import Company
from greenlet import greenlet
from archon.test import LoggedInTestCase, Client, MockWebSocket


save_company = {
    'target': 'company',
    'type': 'save',
    'data': {
        'name': 'Butcher',
    },
}

update_company = {
    'target': 'company',
    'type': 'update',
    'data': {
        'name': 'Hairdresser',
        'id': 1,
    }
}

subscribe_company = {
    'requestId': '1234',
    'target': 'company',
    'data': {
        'name': 'Hairdresser',
    },
    'type': 'subscribe',
}


class TestUpdate(LoggedInTestCase):
    def setUp(self):
        self.client = Client(app, db)
        super().setUp()

    def test_update_model(self):
        ws = MockWebSocket()
        ws.mock_incoming_message(save_company)

        g = greenlet(self.client.open_connection)
        g.switch(ws, app=self.client.app)

        self.assertEqual(1, len(ws.outgoing_messages))
        res_save = json.loads(ws.outgoing_messages[0])
        self.assertEqual(res_save['code'], 'success')

        ws.mock_incoming_message(update_company)
        g.switch()

        self.assertEqual(2, len(ws.outgoing_messages))
        res_update = json.loads(ws.outgoing_messages[1])

        self.assertEqual('update', res_update['type'])
        self.assertEqual('company', res_update['target'])
        self.assertEqual('success', res_update['code'])
        self.assertEqual('Hairdresser', res_update['data']['name'])
        self.assertEqual(1, res_update['data']['id'])

    def test_fails_without_id(self):
        ws = MockWebSocket()
        ws.mock_incoming_message(save_company)

        g = greenlet(self.client.open_connection)
        g.switch(ws, app=self.client.app)

        self.assertEqual(1, len(ws.outgoing_messages))
        res_save = json.loads(ws.outgoing_messages[0])
        self.assertEqual(res_save['code'], 'success')

        update_no_id = {
            'target': 'company',
            'type': 'update',
            'data': {
                'name': 'Hairdresser',
            }
        }

        ws.mock_incoming_message(update_no_id)
        g.switch()

        self.assertEqual(2, len(ws.outgoing_messages))
        res_update = json.loads(ws.outgoing_messages[1])

        self.assertDictEqual({
            'type': 'update',
            'code': 'error',
            'message': 'No id given',
        }, res_update)


class TestSub(LoggedInTestCase):
    def setUp(self):
        self.client = Client(app, db)
        super().setUp()

    def test_update_into_scope(self):
        ws = MockWebSocket()

        c = Company({
                'name': 'Butcher',
            })
        db.session.add(c)
        db.session.commit()

        ws.mock_incoming_message(subscribe_company)
        ws.mock_incoming_message(update_company)

        self.client.open_connection(ws)

        self.assertEqual(3, len(ws.outgoing_messages))
        r1 = json.loads(ws.outgoing_messages[0])
        r2 = json.loads(ws.outgoing_messages[1])
        r3 = json.loads(ws.outgoing_messages[2])

        self.assertEqual('publish', r1['type'])
        self.assertEqual(0, len(r1['data']['add']))
        self.assertEqual('success', r1['code'])

        self.assertEqual('update', r3['type'])
        self.assertEqual('success', r3['code'])

        self.assertDictEqual({
            'type': 'publish',
            'target': 'company',
            'requestId': '1234',
            'data': {
                'add': [{
                    'id': c.id,
                    'name': c.name,
                }]
            },
        }, r2)

    def test_update_out_of_scope(self):
        ws = MockWebSocket()

        c = Company({
                'name': 'Butcher',
            })
        db.session.add(c)
        db.session.commit()

        subscribe_company = {
            'requestId': '9012',
            'target': 'company',
            'data': {
                'name': 'Butcher',
            },
            'type': 'subscribe',
        }

        ws.mock_incoming_message(subscribe_company)
        ws.mock_incoming_message(update_company)

        self.client.open_connection(ws)

        self.assertEqual(3, len(ws.outgoing_messages))
        r1 = json.loads(ws.outgoing_messages[0])
        r2 = json.loads(ws.outgoing_messages[1])
        r3 = json.loads(ws.outgoing_messages[2])

        self.assertEqual('publish', r1['type'])
        self.assertEqual(1, len(r1['data']['add']))
        self.assertEqual('success', r1['code'])

        self.assertEqual('update', r3['type'])
        self.assertEqual('success', r3['code'])

        self.assertDictEqual({
            'type': 'publish',
            'target': 'company',
            'requestId': '9012',
            'data': {
                'remove': [{
                    'id': c.id,
                    'name': c.name,
                }]
            },
        }, r2)

    def test_update_already_in_scope(self):
        ws = MockWebSocket()

        c = Company({
                'name': 'Butcher',
            })
        db.session.add(c)
        db.session.commit()

        subscribe_company = {
            'requestId': '5678',
            'target': 'company',
            'data': {
                'id': 1,
            },
            'type': 'subscribe',
        }

        ws.mock_incoming_message(subscribe_company)
        ws.mock_incoming_message(update_company)

        self.client.open_connection(ws)

        self.assertEqual(3, len(ws.outgoing_messages))
        r1 = json.loads(ws.outgoing_messages[0])
        r2 = json.loads(ws.outgoing_messages[1])
        r3 = json.loads(ws.outgoing_messages[2])

        self.assertEqual('publish', r1['type'])
        self.assertEqual(1, len(r1['data']['add']))
        self.assertEqual('success', r1['code'])

        self.assertEqual('update', r3['type'])
        self.assertEqual('success', r3['code'])

        self.assertDictEqual({
            'type': 'publish',
            'target': 'company',
            'requestId': '5678',
            'data': {
                'update': [{
                    'id': c.id,
                    'name': c.name,
                }]
            },
        }, r2)



import json
from .testapp.app import app, db
from .testapp.models import Company, Employee
from archon.test import LoggedInTestCase, TestCase, Client, MockWebSocket


class TestCustom(LoggedInTestCase):
    def setUp(self):
        self.client = Client(app, db)
        super().setUp()

    def test_company(self):
        ws = MockWebSocket()
        ws.mock_incoming_message({
            'target': 'company',
            'type': 'ask_for_raise',
        })

        self.client.open_connection(ws)

        self.assertEqual(1, len(ws.outgoing_messages))
        m = ws.outgoing_messages[0]

        self.assertDictEqual({
            'answer': 'no',
            'code': 'error',
        }, json.loads(m))

    def test_db_connection(self):
        c = Company({'name': 'Post office'})
        db.session.add(c)
        e1 = Employee({'name': 'Henk'})
        e2 = Employee({'name': 'Peter'})
        e1.company = c
        e2.company = c
        db.session.add(e1, e2)
        db.session.commit()

        ws = MockWebSocket()
        ws.mock_incoming_message({
            'target': 'company',
            'data': {'id': 1},
            'type': 'list_employees',
        })
        self.client.open_connection(ws)

        self.assertEqual(1, len(ws.outgoing_messages))
        m = ws.outgoing_messages[0]

        self.assertDictEqual({
            'code': 'success',
            'data': ['Henk', 'Peter'],
            'type': 'list_employees',
        }, json.loads(m))

    def test_exception(self):
        ws = MockWebSocket()
        ws.mock_incoming_message({
            'target': 'company',
            'data': {'id': 1},
            'type': 'list_employees_error',
        })
        self.client.open_connection(ws)

        self.assertEqual(1, len(ws.outgoing_messages))
        m = ws.outgoing_messages[0]

        self.assertDictEqual({
            'code': 'error',
            'type': 'list_employees_error',
            'message': "'CompanyController' object has no attribute 'type'",
            }, json.loads(m))


class TestNotAuth(TestCase):
    def setUp(self):
        self.client = Client(app, db)
        super().setUp()

    def test_company(self):
        ws = MockWebSocket()
        ws.mock_incoming_message({
            'target': 'company',
            'type': 'ask_for_raise',
        })

        self.client.open_connection(ws)

        self.assertEqual(1, len(ws.outgoing_messages))
        m = ws.outgoing_messages[0]

        self.assertDictEqual({
            'code': 'error',
            'message': 'Unauthorized',
            'type': 'ask_for_raise',
        }, json.loads(m))
